@@ -17,6 +17,7 @@ use {
         path::{Path, PathBuf},
         sync::atomic::AtomicU64,
         sync::Arc,
+        time::Instant,
     },
     tokio::sync::oneshot::{self, Sender},
     tracing::{event, info, info_span, warn, Instrument, Level},
@@ -228,7 +229,7 @@ impl ExecuteCtx {
         remote: IpAddr,
     ) -> Result<(), ExecutionError> {
         info!("handling request {} {}", req.method(), req.uri());
-
+        let start_timestamp = Instant::now();
         let session = Session::new(
             req_id,
             req,
@@ -276,11 +277,15 @@ impl ExecuteCtx {
             .expect("`memory` is exported")
             .size(&store);
 
+        let request_duration = Instant::now().duration_since(start_timestamp);
+
         info!(
             "request completed using {} of WebAssembly heap",
             bytesize::ByteSize::kib(heap_pages as u64 * 64)
         );
 
+        info!("request completed in {:?}", request_duration);
+
         outcome
     }
 }
@@ -300,13 +305,15 @@ fn configure_wasmtime() -> wasmtime::Config {
     let module_limits = ModuleLimits {
         // allow for up to 128MiB of linear memory
         memory_pages: 2048,
-        // Default limit on types is 100, but some js programs have hit this.
-        // We may have to go higher at some point.
-        types: 200,
+        // ffmpeg-wasi was the last program to break the types
+        types: 1234,
         // AssemblyScript applications tend to create a fair number of globals
-        globals: 64,
+        globals: 1234,
         // Some applications create a large number of functions, in particular in debug mode
-        functions: 20000,
+        // or applications written in swift.
+        functions: 98765,
+        // And every function can end up in the table
+        table_elements: 98765,
         ..ModuleLimits::default()
     };
 
